/*
 * Real time packet sending client
 *
 * Large portions taken from cyclictest
 *
 */

#define _GNU_SOURCE
#include <errno.h>
#include <error.h>
#include <pthread.h>
#include <sched.h>
#include <signal.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "recv_packet.h"
#include "send_packet.h"
#include "utilities.h"

// Structs

typedef struct thread_stat {
  int nb_cycles;
  uint64_t rtt;
  packet_info_t packet_info;
} thread_stat_t;

typedef struct thread_param {
  int interval;
  int max_cycles;
  int priority;
  int etf_offset;

  thread_stat_t stats;
} thread_param_t;

typedef struct main_param {
  int refresh_rate;
  int verbose;
} main_param_t;

typedef struct network_config {
  size_t tx_buffer_len;
  char ip_address[256];
  char network_if[256];
  int packet_priority;
} network_config_t;

// Static functions

static void process_options(int argc, char *argv[]);
static void do_tsn_task(struct thread_param *param, char *data, uint64_t next_txtime);
static void print_histograms();
static void sighand(int sig_num);

// Static variables

static int64_t histograms[NB_HISTOGRAMS][MAX_HIST_VAL];

static main_param_t main_param;
static thread_param_t *param;
static network_config_t network_config;

static int enable_histograms;
static int enable_affinity;
static int enable_etf;
static int enable_timestamps;

enum TSNTask { SEND_PACKET_TASK,
               RTT_TASK };
static enum TSNTask tsn_task;

struct timespec measures_start;
struct timespec measures_end;

char send_data[MAX_BUFFER_SIZE];

static void help(char *argv[]) {
  printf("Usage: %s -f IF  [-abthgv] [-e ETF_OFFSET] [-d BUF_LEN] [-i USEC] [-l N] [-p PRIO] [-q PACKET_PRIO] [-r USEC]\n\n", argv[0]);
  printf("       -a                    Run the real time thread on CPU1\n");
  printf("       -b                    Measure RTT\n");
  printf("       -d BUF_LEN            Set the length of tx buffer\n");
  printf("       -e ETF_OFFSET         Set a txtime with an offset of ETF_OFFSET us (to be used in an ETF qdisc)\n");
  printf("       -f IF                 Set the network interface to be used\n");
  printf("       -g                    Print histograms to sdtout on exit\n");
  printf("       -h                    Show help\n");
  printf("       -i USEC               Wake up the real time thread every USEC microseconds (Default: 10ms)\n");
  printf("       -l N                  Wake up the real time thread N times (Default: 0)\n");
  printf("       -p PRIO               Run the real time thread at priority PRIO\n");
  printf("       -q PACKET_PRIO        Send packets with PACKET_PRIO priority\n");
  printf("       -r USEC               Refresh the non real time main thread every USEC microseconds (Default: 50ms)\n");
  printf("       -t                    Enable timestamps\n");
  printf("       -v                    Verbose\n");
  printf("\n");
}

// Real-time thread
// Sends packets at a regular intervall
static void *packet_sending_thread(void *p) {
  struct timespec next;
  uint64_t next_txtime;
  struct sched_param priority;
  thread_param_t *param = (thread_param_t *)p;
  thread_stat_t *stats = &param->stats;
  cpu_set_t mask;

  // Set thread CPU affinity
  if (enable_affinity) {
    CPU_ZERO(&mask);
    CPU_SET(1, &mask);
    if (sched_setaffinity(0, sizeof(mask), &mask))
      error(EXIT_FAILURE, errno, "Could not set CPU affinity to CPU #1\n");
  }

  // Set thread priority
  priority.sched_priority = param->priority;
  if (sched_setscheduler(0, SCHED_FIFO, &priority))
    error(EXIT_FAILURE, errno, "Couldn't set priority");

  if (enable_etf) {
    // Measure from CLOCK_TAI to generate timestamp
    clock_gettime(CLOCK_TAI, &next);
    next_txtime = next.tv_sec * NSEC_PER_SEC + next.tv_nsec;
    next_txtime += param->etf_offset;
  } else {
    next_txtime = 0;
  }

  clock_gettime(CLOCK_MONOTONIC, &next);
  clock_gettime(CLOCK_MONOTONIC, &measures_start);
  // Packet sending loop
  for (stats->nb_cycles = 0;; stats->nb_cycles++) {
    if (param->max_cycles)
      if (stats->nb_cycles >= param->max_cycles)
        break;

    sprintf(send_data, "%d", stats->nb_cycles % 1000);
    do_tsn_task(param, send_data, next_txtime);

    add_ns(&next, param->interval);

    if (enable_etf)
      next_txtime += param->interval;

    clock_nanosleep(CLOCK_MONOTONIC, TIMER_ABSTIME, &next, NULL);
  }

  return NULL;
}

// Main thread, has non-real time priority
// Handles the IO and creates real time threads
int main(int argc, char *argv[]) {
  pthread_t thread;
  thread_stat_t *stats;

  param = malloc(sizeof(thread_param_t));
  stats = &param->stats;

  // Default configuration values
  param->interval = 100000 * 1000;
  param->max_cycles = 0;
  param->priority = 99;

  enable_affinity = 0;
  enable_etf = 0;
  enable_timestamps = 0;
  enable_histograms = 0;
  tsn_task = SEND_PACKET_TASK;

  network_config.packet_priority = 3;
  network_config.tx_buffer_len = 1024;

  main_param.refresh_rate = 50000;
  main_param.verbose = 0;

  // Process bash options
  process_options(argc, argv);

  if (enable_histograms) {
    // Init histograms
    memset((int64_t *)histograms, 0, NB_HISTOGRAMS * MAX_HIST_VAL);
  }

  // Catch breaks with sighand to print the histograms
  init_signals(sighand, enable_histograms);

  // Initialize the UDP packet sending socket
  init_udp_send(enable_etf, enable_timestamps,
                network_config.packet_priority,
                network_config.network_if,
                network_config.tx_buffer_len);

  // Initialize the UDP packet receiving socket if RTT is measured
  if (tsn_task == RTT_TASK)
    init_udp_recv(0, network_config.network_if);

  // Create the real time thread
  if (pthread_create(&thread, NULL, packet_sending_thread, (void *)param))
    error(EXIT_FAILURE, errno, "Couldn't create thread");

  // Verbose loop
  for (;;) {
    usleep(main_param.refresh_rate);

    if (main_param.verbose) {
      if (tsn_task == RTT_TASK) {

        printf("%*d:   RTT: %*" PRIu64 "\n", 10, stats->nb_cycles, 10, stats->rtt);
	printf("\033[%dA", 1);
      } else if (enable_timestamps) {

        int64_t user_space_time = stats->packet_info.userspace_exit_ts - stats->packet_info.userspace_enter_ts;
        int64_t kernel_space_time = stats->packet_info.kernelspace_ts - stats->packet_info.userspace_exit_ts;

        printf("%*d:   U: %*" PRIi64 ",   K: %*" PRIi64 "\n",
               10, stats->nb_cycles, 10, user_space_time, 10, kernel_space_time);
	printf("\033[%dA", 1);
      }
    }

    if (param->max_cycles)
      if (param->max_cycles == stats->nb_cycles)
        break;
  }

  if (enable_histograms)
    print_histograms();

  exit(EXIT_SUCCESS);
}

// Critical TSN task
static void do_tsn_task(struct thread_param *param, char *data, uint64_t next_txtime) {
  struct timespec t1, t2;
  int rtt_us;

  // One way packet sending
  if (tsn_task == SEND_PACKET_TASK) {

    param->stats.packet_info = send_udp_packet(
        enable_etf, enable_timestamps, data, next_txtime,
        network_config.ip_address, histograms);

    // Round Trip Time measurement
  } else if (tsn_task == RTT_TASK) {

    clock_gettime(CLOCK_MONOTONIC, &t1);
    send_udp_packet(0, 0, data, next_txtime,
                    network_config.ip_address, NULL);
    recv_udp_packet(0, 0, NULL);
    clock_gettime(CLOCK_MONOTONIC, &t2);

    param->stats.rtt = calcdiff_ns(t2, t1);

    if (enable_histograms) {
      rtt_us = param->stats.rtt / 1000;

      if (rtt_us > MAX_HIST_VAL)
        fprintf(stderr, "RTT value higher than MAX_HIST_VAL : %d ( > %d)\n", rtt_us, MAX_HIST_VAL);
      else
        histograms[0][rtt_us]++;
    }
  }
}

// Print histograms in .json format
static void print_histograms() {

  uint64_t duration;
  int duration_hour, duration_minutes, interval;
  int max_hist_val, nb_hists;

  clock_gettime(CLOCK_MONOTONIC, &measures_end);

  duration = calcdiff_ns(measures_end, measures_start);
  duration_hour = duration / NSEC_PER_SEC / 3600;
  duration_minutes = duration / NSEC_PER_SEC / 60 - duration_hour * 60;

  interval = param->interval / 1000;

  if (tsn_task == SEND_PACKET_TASK) {
    printf("{\"measure_sets\": [{"
           "\"measure_type\": \"packet_send_timestamps\","
           "\"props_names\": [\"user_space\", \"kernel_space\"],"
           "\"units\": [\"us\", \"us\"],"
           "\"props_type\": \"histogram\","
           "\"metadata\": {"
           "\"i\": \"%dus\", \"duration\": \"%dh%d\", \"etf_offset\": \"%dus\","
           "},"
           "\"props\": [",
           interval, duration_hour, duration_minutes, param->etf_offset);
  } else if (tsn_task == RTT_TASK) {
    printf("{\"measure_sets\": [{"
           "\"measure_type\": \"packet_rtt\","
           "\"props_names\": [\"rtt\"],"
           "\"units\": [\"us\"],"
           "\"props_type\": \"histogram\","
           "\"metadata\": {"
           "\"i\": \"%dus\", \"duration\": \"%dh%d\", \"etf_offset\": \"%dus\","
           "},"
           "\"props\": [",
           interval, duration_hour, duration_minutes, param->etf_offset);
  }

  nb_hists = tsn_task == SEND_PACKET_TASK ? 2 : 1;

  max_hist_val = 0;
  for (int i = 0; i < nb_hists; i++)
    for (int j = 0; j < MAX_HIST_VAL; j++)
      if (histograms[i][j])
        max_hist_val = j > max_hist_val ? j : max_hist_val;

  for (int i = 0; i < nb_hists; i++) {
    printf("[");
    for (int j = 0; j < max_hist_val; j++)
      printf("%" PRIi64 "%s", histograms[i][j], (j + 1 < max_hist_val ? ", " : ""));
    printf("%s", (i + 1 < nb_hists ? "], " : "]"));
  }

  printf("]}]}\n");
}

static void sighand(int sig_num) {
  (void)sig_num;
  print_histograms();
  exit(EXIT_SUCCESS);
}

// Process bash options
static void process_options(int argc, char *argv[]) {

  int network_if_specified = 0;

  for (;;) {
    int c = getopt(argc, argv, "abd:e:f:ghi:l:p:q:r:tv");

    if (c == -1)
      break;

    switch (c) {
    case 'a':
      enable_affinity = 1;
      break;
    case 'b':
      tsn_task = RTT_TASK;
      break;
    case 'd':
      network_config.tx_buffer_len = atoi(optarg);
      if (network_config.tx_buffer_len < 1) {
        fprintf(stderr, "BUF_LEN should be greater than 1\n");
        exit(EXIT_FAILURE);
      }
      break;
    case 'e':
      enable_etf = 1;
      param->etf_offset = atoi(optarg) * 1000;
      break;
    case 'f':
      network_if_specified = 1;
      strcpy(network_config.network_if, optarg);
      break;
    case 'g':
      enable_histograms = 1;
      break;
    case 'h':
      help(argv);
      exit(EXIT_SUCCESS);
      break;
    case 'i':
      param->interval = atoi(optarg) * 1000;
      break;
    case 'l':
      param->max_cycles = atoi(optarg);
      break;
    case 'p':
      param->priority = atoi(optarg);
      break;
    case 'q':
      network_config.packet_priority = atoi(optarg);
      break;
    case 'r':
      main_param.refresh_rate = atoi(optarg);
      break;
    case 't':
      enable_timestamps = 1;
      break;
    case 'v':
      main_param.verbose = 1;
      break;
    }
  }

  if (!network_if_specified) {
    fprintf(stderr, "You need to specifiy an network interface\n");
    help(argv);
    exit(EXIT_FAILURE);
  }

  if (argc != optind + 1) {
    if (argc < optind + 1)
      fprintf(stderr, "You need to specifiy an IP address\n");
    else
      fprintf(stderr, "Too many arguments\n");
    help(argv);
    exit(EXIT_FAILURE);
  }
  strcpy(network_config.ip_address, argv[optind]);
}
