/*
 *
 * UDP packet sending functions
 *
 * Large portions taken from scheduled tx tools gist
 *
 */

#define _GNU_SOURCE
#include <arpa/inet.h>
#include <errno.h>
#include <error.h>
#include <fcntl.h>
#include <ifaddrs.h>
#include <inttypes.h>
#include <linux/errqueue.h>
#include <linux/ethtool.h>
#include <linux/net_tstamp.h>
#include <linux/sockios.h>
#include <net/if.h>
#include <netdb.h>
#include <netinet/in.h>
#include <poll.h>
#include <pthread.h>
#include <sched.h>
#include <signal.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include "send_packet.h"
#include "utilities.h"

static void process_timestamps(packet_info_t *packet_info, int64_t histograms[NB_HISTOGRAMS][MAX_HIST_VAL]);
static void init_tx_buffer(size_t _tx_buffer_len);

static int so_priority = 3;
static struct sock_txtime sk_txtime;
static char *tx_buffer;
static size_t tx_buffer_len;
static int sock_fd;

static int64_t tai_offset;

static int so_timestamping_flags =
    SOF_TIMESTAMPING_TX_SOFTWARE | SOF_TIMESTAMPING_SOFTWARE;

// Sets the interface
static int set_if(char *network_if) {
  struct ifreq ifreq;

  memset(&ifreq, 0, sizeof(ifreq));
  strncpy(ifreq.ifr_name, network_if, sizeof(ifreq.ifr_name) - 1);

  if (ioctl(sock_fd, SIOCGIFINDEX, &ifreq))
    error(EXIT_FAILURE, errno, "ioctl SIOCGIFINDEX failed\n");

  return ifreq.ifr_ifindex;
}

static void init_tx_buffer(size_t _tx_buffer_len) {
  if (_tx_buffer_len < 1) {
    fprintf(stderr, "tx buffer length should be greater than 1\n");
    exit(EXIT_FAILURE);
  }

  tx_buffer_len = _tx_buffer_len;
  tx_buffer = malloc(tx_buffer_len);
}

/*
 * Init UDP socket
 */
void init_udp_send(int use_etf, int use_timestamps, int packet_priority,
                   char *network_if, size_t _tx_buffer_len) {
  int set_if_err;

  struct timespec ts_mon;
  struct timespec ts_tai;

  init_tx_buffer(_tx_buffer_len);

  clock_gettime(CLOCK_MONOTONIC, &ts_mon);
  clock_gettime(CLOCK_TAI, &ts_tai);
  tai_offset = (ts_mon.tv_sec - ts_tai.tv_sec) * NSEC_PER_SEC +
               (ts_mon.tv_nsec - ts_tai.tv_nsec);

  so_priority = packet_priority;

  sock_fd = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
  if (sock_fd < 0)
    error(EXIT_FAILURE, errno, "Socket creation failed\n");

  set_if_err = set_if(network_if);
  if (set_if_err < 0)
    error(EXIT_FAILURE, errno, "Couldn't set interface\n");

  if (setsockopt(sock_fd, SOL_SOCKET, SO_PRIORITY, &so_priority,
                 sizeof(so_priority)))
    error(EXIT_FAILURE, errno, "Couldn't set socket priority\n");

  if (setsockopt(sock_fd, SOL_SOCKET, SO_BINDTODEVICE, network_if,
                 strlen(network_if)))
    error(EXIT_FAILURE, errno, "setsockopt SO_BINDTODEVICE failed\n");

  if (use_etf) {
    sk_txtime.clockid = CLOCK_TAI;
    sk_txtime.flags = 0;

    if (setsockopt(sock_fd, SOL_SOCKET, SO_TXTIME, &sk_txtime, sizeof(sk_txtime)))
      error(EXIT_FAILURE, errno, "setsockopt SO_TXTIME failed\n");
  }

  if (use_timestamps) {
    if (setsockopt(sock_fd, SOL_SOCKET, SO_TIMESTAMPING, &so_timestamping_flags,
                   sizeof(so_timestamping_flags)))
      error(EXIT_FAILURE, errno, "setsockopt SO_TIMESTAMPING failed\n");
  }
}

/*
 * Sends udp packets
 */
packet_info_t send_udp_packet(int use_etf, int use_timestamps,
                              char *data,
                              uint64_t txtime,
                              const char *server_ip,
                              int64_t histograms[NB_HISTOGRAMS][MAX_HIST_VAL]) {

  struct msghdr msg;                             // Message hardware, sent to the socket
  struct cmsghdr *cmsg;                          // Control message hardware, for txtime
  char control[CMSG_SPACE(sizeof(txtime))] = {}; // Stores txtime
  struct iovec iov;                              // The iovec structures stores the TX buffer
  // Poll file descriptor, used to poll for timestamp messages
  struct pollfd poll_fd = {sock_fd, POLLPRI, 0};
  int sendmsgerr, pollerr;
  // Server address
  struct sockaddr_in sin;

  packet_info_t packet_info;
  struct timespec ts;

  if (use_timestamps) {
    clock_gettime(CLOCK_REALTIME, &ts);
    packet_info.userspace_enter_ts = ts_to_uint(ts);
  }

  strcpy(tx_buffer, data);

  memset(&sin, 0, sizeof(sin));
  sin.sin_family = AF_INET;
  sin.sin_addr.s_addr = inet_addr(server_ip);
  sin.sin_port = htons(SERVER_PORT_INT);

  iov.iov_base = tx_buffer;
  iov.iov_len = tx_buffer_len;

  memset(&msg, 0, sizeof(msg));
  msg.msg_name = &sin;
  msg.msg_namelen = sizeof(sin);
  msg.msg_iov = &iov;
  msg.msg_iovlen = 1;

  if (use_etf) {
    // We specify the transmission time in the CMSG.
    msg.msg_control = control;
    msg.msg_controllen = sizeof(control);

    cmsg = CMSG_FIRSTHDR(&msg);
    cmsg->cmsg_level = SOL_SOCKET;
    cmsg->cmsg_type = SCM_TXTIME;
    cmsg->cmsg_len = CMSG_LEN(sizeof(uint64_t));
    *((uint64_t *)CMSG_DATA(cmsg)) = txtime;
    msg.msg_controllen = cmsg->cmsg_len;
  }

  if (use_timestamps) {
    clock_gettime(CLOCK_REALTIME, &ts);
    packet_info.userspace_exit_ts = ts_to_uint(ts);
  }

  sendmsgerr = sendmsg(sock_fd, &msg, 0);
  if (sendmsgerr < 0)
    error(EXIT_FAILURE, errno, "sendmsg failed, ret value: %d\n", sendmsgerr);

  if (use_timestamps) {
    pollerr = poll(&poll_fd, 1, 1);
    if (pollerr > 0)
      process_timestamps(&packet_info, histograms);
    else
      fprintf(stderr, "select failed\n");
  }

  return packet_info;
}

static void fill_histograms(packet_info_t *packet_info, int64_t histograms[NB_HISTOGRAMS][MAX_HIST_VAL]) {

  uint64_t user_space_time = packet_info->userspace_exit_ts - packet_info->userspace_enter_ts;
  uint64_t kernel_space_time = packet_info->kernelspace_ts - packet_info->userspace_exit_ts;

  user_space_time /= 1000u;
  kernel_space_time /= 1000u;

  if (user_space_time > MAX_HIST_VAL)
    fprintf(stderr, "user_space_time value too high: %" PRIu64 "us\n", user_space_time);
  else
    histograms[0][user_space_time]++;

  if (kernel_space_time > MAX_HIST_VAL)
    fprintf(stderr, "kernel_space_time value too high: %" PRIu64 "us\n", kernel_space_time);
  else
    histograms[1][kernel_space_time]++;
}

static void process_timestamps(packet_info_t *packet_info, int64_t histograms[NB_HISTOGRAMS][MAX_HIST_VAL]) {
  char data[256];
  struct msghdr msg;
  struct iovec entry;
  struct sockaddr_in from_addr;
  struct {
    struct cmsghdr cm;
    char control[512];
  } control;
  struct cmsghdr *cmsg;

  memset(&msg, 0, sizeof(msg));
  msg.msg_iov = &entry;
  msg.msg_iovlen = 1;
  entry.iov_base = data;
  entry.iov_len = sizeof(data);
  msg.msg_name = (caddr_t)&from_addr;
  msg.msg_namelen = sizeof(from_addr);
  msg.msg_control = &control;
  msg.msg_controllen = sizeof(control);

  if (recvmsg(sock_fd, &msg, MSG_ERRQUEUE | MSG_DONTWAIT) == -1) {
    fprintf(stderr, "recvmsg failed\n");
    return;
  }

  for (cmsg = CMSG_FIRSTHDR(&msg); cmsg; cmsg = CMSG_NXTHDR(&msg, cmsg)) {
    if (cmsg->cmsg_level == SOL_SOCKET && cmsg->cmsg_type == SO_TIMESTAMPING) {

      struct timespec *stamp = (struct timespec *)CMSG_DATA(cmsg);
      packet_info->kernelspace_ts = ts_to_uint(*stamp);
      fill_histograms(packet_info, histograms);

    } else {

#ifdef DEBUG
      fprintf(stderr, "process_timestamps: level %d type %d", cmsg->cmsg_level,
              cmsg->cmsg_type);
#endif
    }
  }
}

#ifdef DEBUG

/*
 * Code from scheduled_tx_tools
 */
static int process_socket_error_queue() {
  uint8_t msg_control[CMSG_SPACE(sizeof(struct sock_extended_err))];
  unsigned char err_buffer[256];
  struct sock_extended_err *serr;
  struct cmsghdr *cmsg;
  __u64 tstamp = 0;

  struct iovec iov = {.iov_base = err_buffer, .iov_len = sizeof(err_buffer)};
  struct msghdr msg = {.msg_iov = &iov,
                       .msg_iovlen = 1,
                       .msg_control = msg_control,
                       .msg_controllen = sizeof(msg_control)};

  if (recvmsg(sock_fd, &msg, MSG_ERRQUEUE) == -1) {
    fprintf(stderr, "recvmsg failed");
    return -1;
  }

  cmsg = CMSG_FIRSTHDR(&msg);
  while (cmsg != NULL) {
    serr = (void *)CMSG_DATA(cmsg);
    if (serr->ee_origin == SO_EE_ORIGIN_TXTIME) {
      tstamp = ((__u64)serr->ee_data << 32) + serr->ee_info;

      switch (serr->ee_code) {
      case SO_EE_CODE_TXTIME_INVALID_PARAM:
        fprintf(stderr,
                "packet with tstamp %llu dropped due to invalid params\n",
                tstamp);
        return 0;
      case SO_EE_CODE_TXTIME_MISSED:
        fprintf(stderr,
                "packet with tstamp %llu dropped due to missed deadline\n",
                tstamp);
        return 0;
      default:
        return -1;
      }
    }

    cmsg = CMSG_NXTHDR(&msg, cmsg);
  }

  return 0;
}
#endif
