/*
 * Real time packet receiving server
 *
 * Large portions taken from cyclictest
 *
 */

#define _GNU_SOURCE
#include <arpa/inet.h>
#include <errno.h>
#include <error.h>
#include <inttypes.h>
#include <netdb.h>
#include <netinet/in.h>
#include <pthread.h>
#include <sched.h>
#include <signal.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>

#include "recv_packet.h"
#include "send_packet.h"
#include "utilities.h"

// Structs

typedef struct thread_stat {
  uint64_t min_interval;
  uint64_t max_interval;
  int packets_received;
  packet_info_t packet_info;
  int lost_packets;
} thread_stat_t;

typedef struct thread_param {
  int interval;
  int priority;

  thread_stat_t stats;
} thread_param_t;

typedef struct network_config {
  size_t tx_buffer_len;
  char ip_address[256];
  char network_if[256];
  int packet_priority;
} network_config_t;

typedef struct main_param {
  int refresh_rate;
  int verbose;
  size_t tx_buffer_len;
} main_param_t;

static void process_options(int argc, char *argv[]);
static void print_histograms();
static void sighand(int sig_num);

// Static variables

static int64_t histograms[NB_HISTOGRAMS][MAX_HIST_VAL];

static main_param_t main_param;
static thread_param_t *param;
static network_config_t network_config;

static int enable_histograms;
static int enable_affinity;
static int enable_timestamps;

enum TSNTask { RECV_PACKET_TASK,
               RTT_TASK,
               XDP_TASK };
static enum TSNTask tsn_task;

struct timespec measures_start;
struct timespec measures_end;

static void help(char *argv[]) {
  printf("Usage: %s [-aghtv] [-b CLIENT_IP] [-d BUF_LEN] [-f IF] [-i USEC] [-p PRIO] [-r USEC] [-x]\n\n", argv[0]);
  printf("       -a                    Run the real time thread on CPU1\n");
  printf("       -b CLIENT_IP          Server side RTT\n");
  printf("       -d BUF_LEN            Set the length of tx buffer\n");
  printf("       -f IF                 Set the network interface to be used\n");
  printf("       -g                    Print histograms to sdtout on exit\n");
  printf("       -h                    Show help\n");
  printf("       -i USEC               Wake up the real time thread every USEC microseconds (Default: 10ms)\n");
  printf("       -p PRIO               Run the real time thread at priority PRIO\n");
  printf("       -r USEC               Refresh the non real time main thread every USEC microseconds\n");
  printf("       -t                    Enable timestamps\n");
  printf("       -v                    Verbose\n");
  printf("       -x                    Use AF_XDP sockets\n");
  printf("\n");
}

// Real-time thread
// Measures intervals between packet receptions
static void *packet_receiving_thread(void *p) {
  struct timespec current, previous;
  struct sched_param priority;
  thread_param_t *param = (thread_param_t *)p;
  thread_stat_t *stats = &param->stats;
  uint64_t diff = 0;
  cpu_set_t mask;
  int64_t dist_to_interval;
  int prev_packet_id = 0;

  stats->min_interval = UINT64_MAX;
  stats->max_interval = 0;

  if (enable_affinity) {
    // Set thread CPU affinity
    CPU_ZERO(&mask);
    CPU_SET(1, &mask);
    if (sched_setaffinity(0, sizeof(mask), &mask))
      error(EXIT_FAILURE, errno, "Could not set CPU affinity to CPU #1\n");
  }

  // Set thread priority
  priority.sched_priority = param->priority;
  if (sched_setscheduler(0, SCHED_FIFO, &priority))
    error(EXIT_FAILURE, errno, "Couldn't set priority");

  clock_gettime(CLOCK_MONOTONIC, &measures_start);
  // Packet receiving loop
  for (stats->packets_received = 0;; stats->packets_received++) {

    if (tsn_task == RTT_TASK) {

      recv_udp_packet(0, 0, NULL);
      send_udp_packet(0, 0, "", 0, network_config.ip_address, NULL);

    } else if (tsn_task == RECV_PACKET_TASK) {

      int current_packet_id;

      param->stats.packet_info = recv_udp_packet(enable_timestamps, enable_histograms, histograms);

      clock_gettime(CLOCK_MONOTONIC, &current);

      current_packet_id = atoi(param->stats.packet_info.data);

      // If this is not the first received packet
      if (stats->packets_received) {

        diff = calcdiff_ns(current, previous);
        stats->min_interval = diff < stats->min_interval ? diff : stats->min_interval;
        stats->max_interval = diff > stats->max_interval ? diff : stats->max_interval;

        // Check if packets were lost
        param->stats.lost_packets += (current_packet_id - prev_packet_id - 1) % 1000;

        if (enable_histograms) {
          dist_to_interval = (((int64_t)diff) - param->interval) / 1000;
          dist_to_interval += MAX_HIST_VAL / 2;

          if (dist_to_interval > ((int)MAX_HIST_VAL) || dist_to_interval < 0)
            fprintf(stderr, "jitter higher than MAX_HIST_VAL: %" PRIi64 "\n", dist_to_interval);
          else
            histograms[2][dist_to_interval]++;
        }
      }

      previous = current;
      prev_packet_id = current_packet_id;
    } else if (tsn_task == XDP_TASK) {

      if(recv_xdp_packet(&current))
        continue;

      if (stats->packets_received) {

        diff = calcdiff_ns(current, previous);
        stats->min_interval = diff < stats->min_interval ? diff : stats->min_interval;
        stats->max_interval = diff > stats->max_interval ? diff : stats->max_interval;

        if (enable_histograms) {
          dist_to_interval = (((int64_t)diff) - param->interval) / 1000;
          dist_to_interval += MAX_HIST_VAL / 2;

          if (dist_to_interval > ((int)MAX_HIST_VAL) || dist_to_interval < 0)
            fprintf(stderr, "jitter higher than MAX_HIST_VAL: %" PRIi64 "\n", dist_to_interval);
          else
            histograms[2][dist_to_interval]++;
        }
      }

      previous = current;
    }
  }

  return NULL;
}

// Main thread, has non-real time priority
// Handles the IO and creates real time threads
int main(int argc, char *argv[]) {
  pthread_t thread;
  thread_stat_t *stats;

  param = malloc(sizeof(thread_param_t));
  stats = &param->stats;

  // Default configuration values
  param->interval = 100000 * 1000;
  param->priority = 99;

  enable_affinity = 0;
  enable_timestamps = 0;
  enable_histograms = 0;
  tsn_task = RECV_PACKET_TASK;

  network_config.tx_buffer_len = 1024;

  main_param.refresh_rate = 50000;
  main_param.verbose = 0;

  // Process bash options
  process_options(argc, argv);

  if (enable_histograms) {
    // Init histograms
    memset((int64_t *)histograms, 0, NB_HISTOGRAMS * MAX_HIST_VAL);
  }

  // Catch breaks with sighand to print the histograms
  init_signals(sighand, enable_histograms);

  if (tsn_task == XDP_TASK) {

    // Initialize the XDP packet receiving socket
    init_xdp_recv(network_config.network_if);

  } else {

    // Initialize the UDP packet receiving socket
    init_udp_recv(enable_timestamps, network_config.network_if);

    // Initialize the UDP packet sending socket if RTT is measured
    if (tsn_task == RTT_TASK)
      init_udp_send(0, 0, 1, network_config.network_if, network_config.tx_buffer_len);
  }

  // Create the real time thread
  if (pthread_create(&thread, NULL, packet_receiving_thread, (void *)param))
    error(EXIT_FAILURE, errno, "Couldn't create thread");

  // Verbose loop
  for (;;) {
    usleep(main_param.refresh_rate);

    if (main_param.verbose) {

      if (tsn_task == RECV_PACKET_TASK) {

        int64_t jitter = ((int64_t)stats->max_interval) - stats->min_interval;

        printf("%*d:   J: %*" PRIi64,
               10, stats->packets_received,
               10, jitter);

        if (enable_timestamps) {

          int64_t user_space_time = stats->packet_info.userspace_exit_ts - stats->packet_info.userspace_enter_ts;
          int64_t kernel_space_time = stats->packet_info.userspace_enter_ts - stats->packet_info.kernelspace_ts;

          printf(",   U: %*" PRIi64 ",   K: %*" PRIi64 ",   D: %*s,   L: %*d\n",
                 10, user_space_time,
                 10, kernel_space_time,
                 4, stats->packet_info.data,
                 4, stats->lost_packets);
        } else {
          printf("\n");
        }

        printf("\033[%dA", 1);
      }
    }
  }
}

// Print histograms in .json format
static void print_histograms() {

  uint64_t duration;
  int duration_hour, duration_minutes, interval;
  int max_hist_val, min_hist_val;

  clock_gettime(CLOCK_MONOTONIC, &measures_end);

  duration = calcdiff_ns(measures_end, measures_start);
  duration_hour = duration / NSEC_PER_SEC / 3600;
  duration_minutes = duration / NSEC_PER_SEC / 60 - duration_hour * 60;

  interval = param->interval / 1000;

  if (enable_timestamps) {
    printf("{\"measure_sets\": [{"
           "\"measure_type\": \"packet_recv_timestamps\","
           "\"props_names\": [\"user_space\", \"kernel_space\"],"
           "\"units\": [\"us\", \"us\"],"
           "\"props_type\": \"histogram\","
           "\"metadata\": {"
           "\"i\": \"%dus\", \"duration\": \"%dh%d\""
           "},"
           "\"props\": [",
           interval, duration_hour, duration_minutes);

    max_hist_val = 0;
    for (int i = 0; i < 2; i++)
      for (int j = 0; j < MAX_HIST_VAL; j++)
        if (histograms[i][j])
          max_hist_val = j > max_hist_val ? j : max_hist_val;

    for (int i = 0; i < 2; i++) {
      printf("[");
      for (int j = 0; j < max_hist_val; j++)
        printf("%" PRIi64 "%s", histograms[i][j], (j + 1 < max_hist_val ? ", " : ""));
      printf("%s", (i + 1 < 2 ? "], " : "]"));
    }
  }

  max_hist_val = 0;
  for (int j = 0; j < MAX_HIST_VAL; j++)
    if (histograms[2][j])
      max_hist_val = j;
  min_hist_val = MAX_HIST_VAL - 1;
  for (int j = MAX_HIST_VAL - 1; j >= 0; j--)
    if (histograms[2][j])
      min_hist_val = j;

  if (!enable_timestamps)
    printf("{\"measure_sets\": [{");
  else
    printf("]}, {");

  printf("\"measure_type\": \"packet_jitter\","
         "\"props_names\": [\"jitter\"],"
         "\"units\": [\"us\"],"
         "\"props_type\": \"histogram\","
         "\"middle\": \"%d\","
         "\"metadata\": {"
         "\"i\": \"%dus\", \"duration\": \"%dh%d\""
         "},"
         "\"props\": [[",
         MAX_HIST_VAL / 2 - min_hist_val,
         interval,
         duration_hour,
         duration_minutes);

  for (int j = min_hist_val; j < max_hist_val; j++)
    printf("%" PRIi64 "%s", histograms[2][j], (j + 1 < max_hist_val ? ", " : ""));

  printf("]]}]}\n");
}

static void sighand(int sig_num) {
  (void)sig_num;

  print_histograms();

  if (param->stats.lost_packets)
    fprintf(stderr, "%d packets were lost\n", param->stats.lost_packets);

  exit(EXIT_SUCCESS);
}

// Process bash options
static void process_options(int argc, char *argv[]) {

  int network_if_specified = 0;

  for (;;) {
    int c = getopt(argc, argv, "ab:d:f:ghi:p:r:tvx");

    if (c == -1)
      break;

    switch (c) {
    case 'a':
      enable_affinity = 1;
      break;
    case 'b':
      tsn_task = RTT_TASK;
      strcpy(network_config.ip_address, optarg);
      break;
    case 'd':
      network_config.tx_buffer_len = atoi(optarg);
      if (network_config.tx_buffer_len < 1) {
        fprintf(stderr, "BUF_LEN should be greater than 1\n");
        exit(EXIT_FAILURE);
      }
      break;
    case 'f':
      network_if_specified = 1;
      strcpy(network_config.network_if, optarg);
      break;
    case 'h':
      help(argv);
      exit(EXIT_SUCCESS);
      break;
    case 'i':
      param->interval = atoi(optarg) * 1000;
      break;
    case 'g':
      enable_histograms = 1;
      break;
    case 'p':
      param->priority = atoi(optarg);
      break;
    case 'r':
      main_param.refresh_rate = atoi(optarg);
      break;
    case 't':
      enable_timestamps = 1;
      break;
    case 'v':
      main_param.verbose = 1;
      break;
    case 'x':
      tsn_task = XDP_TASK;
      break;
    }
  }

  if (!network_if_specified) {
    fprintf(stderr, "You need to specifiy an network interface\n");
    help(argv);
    exit(EXIT_FAILURE);
  }

  if (argc != optind) {
    fprintf(stderr, "Too many arguments\n");
    help(argv);
    exit(EXIT_FAILURE);
  }
}
